@@ -22,17 +22,17 @@
 #
 
 import commands
-try:
-    import hashlib as md5
-except ImportError:
-    import md5
 import os
 import shutil
 import subprocess
 import tempfile
 import urlparse
 
-from debian_bundle.changelog import Version
+try:
+    from debian.changelog import Version
+except ImportError:
+    # Prior to 0.1.15 the debian module was called debian_bundle
+    from debian_bundle.changelog import Version
 
 from bzrlib import (
     urlutils,
@@ -75,21 +75,31 @@ from bzrlib.plugins.builddeb.source_distiller import (
         MergeModeDistiller,
         NativeSourceDistiller,
         )
+from bzrlib.plugins.builddeb.tagging import (
+        is_upstream_tag,
+        upstream_tag_version,
+        )
 from bzrlib.plugins.builddeb.upstream import (
+        AptSource,
+        GetOrigSourceSource,
+        PristineTarSource,
+        SelfSplitSource,
+        UScanSource,
         UpstreamProvider,
         UpstreamBranchSource,
-        get_upstream_sources,
         )
 from bzrlib.plugins.builddeb.util import (
         debuild_config,
         dget_changes,
         find_changelog,
         find_last_distribution,
+        find_previous_upload,
         get_export_upstream_revision,
         lookup_distribution,
         open_file,
         open_file_via_transport,
         tarball_name,
+        tree_contains_upstream_source,
         )
 
 dont_purge_opt = Option('dont-purge',
@@ -115,9 +125,6 @@ export_upstream_opt = Option('export-upstream',
 export_upstream_revision_opt = Option('export-upstream-revision',
     help="Select the upstream revision that will be exported.",
     type=str)
-no_user_conf_opt = Option('no-user-config',
-    help="Stop builddeb from reading the user's config file. Used mainly "
-    "for tests.")
 
 
 class cmd_builddeb(Command):
@@ -178,6 +185,9 @@ class cmd_builddeb(Command):
                         short_name='S')
     result_compat_opt = Option('result', help="Present only for compatibility "
             "with bzr-builddeb <= 2.0. Use --result-dir instead.")
+    package_merge_opt = Option('package-merge', help="Build using the "
+            "appropriate -v and -sa options for merging in the changes from "
+            "another source.")
     takes_args = ['branch_or_build_options*']
     aliases = ['bd']
     takes_options = [working_tree_opt, export_only_opt,
@@ -186,7 +196,7 @@ class cmd_builddeb(Command):
         export_upstream_opt, export_upstream_revision_opt,
         quick_opt, reuse_opt, native_opt,
         source_opt, 'revision',
-        no_user_conf_opt, result_compat_opt]
+        result_compat_opt, package_merge_opt]
 
     def _get_tree_and_branch(self, location):
         if location is None:
@@ -298,9 +308,6 @@ class cmd_builddeb(Command):
                 upstream_branch = Branch.open(export_upstream)
                 upstream_branch.lock_read()
                 try:
-                    if export_upstream_revision is None:
-                        export_upstream_revision = \
-                               config.export_upstream_revision
                     if export_upstream_revision is None:
                         export_upstream_revision = \
                                     get_export_upstream_revision(config,
@@ -324,7 +331,7 @@ class cmd_builddeb(Command):
             export_upstream=None, export_upstream_revision=None,
             orig_dir=None, split=None,
             quick=False, reuse=False, native=False,
-            source=False, revision=None, no_user_config=False, result=None):
+            source=False, revision=None, result=None, package_merge=None):
         if result is not None:
             warning("--result is deprected, use --result-dir instead")
         branch, build_options, source = self._branch_and_build_options(
@@ -339,32 +346,54 @@ class cmd_builddeb(Command):
 
         tree.lock_read()
         try:
-            config = debuild_config(tree, working_tree, no_user_config)
+            config = debuild_config(tree, working_tree)
             if reuse:
                 note("Reusing existing build dir")
                 dont_purge = True
                 use_existing = True
             merge, native, split = self._build_type(config, merge, native, split)
             if (not merge and not native and not split and
-                tree.inventory.root.children.keys() == ["debian"]):
+                not tree_contains_upstream_source(tree)):
                 # Default to merge mode if there's only a debian/ directory
                 merge = True
+            (changelog, larstiq) = find_changelog(tree, merge)
+            if package_merge:
+                prev_version = find_previous_upload(tree, merge)
+                build_options.append("-v%s" % str(prev_version))
+                if (prev_version.upstream_version
+                        != changelog.version.upstream_version
+                        or prev_version.epoch != changelog.version.epoch):
+                    build_options.append("-sa")
             build_cmd = self._get_build_command(config, builder, quick,
                     build_options)
-            (changelog, larstiq) = find_changelog(tree, merge)
             result_dir, build_dir, orig_dir = self._get_dirs(config, branch,
                     is_local, result_dir or result, build_dir, orig_dir)
 
-            upstream_branch, upstream_revision = \
-                    self._get_upstream_branch(merge, export_upstream,
-                            export_upstream_revision, config,
-                            changelog.version)
-
-            upstream_provider = UpstreamProvider(
-                    changelog.package, changelog.version,
-                    orig_dir, get_upstream_sources(tree, branch,
-                    larstiq=larstiq, upstream_branch=upstream_branch,
-                    upstream_revision=upstream_revision, allow_split=split))
+            upstream_sources = [
+                PristineTarSource(tree, branch), 
+                AptSource(),
+                ]
+            if merge:
+                upstream_branch, upstream_revision = self._get_upstream_branch(
+                    merge, export_upstream, export_upstream_revision, config,
+                    changelog.version)
+                if upstream_branch is not None:
+                    upstream_sources.append(UpstreamBranchSource(
+                            upstream_branch,
+                            {changelog.version.upstream_version:
+                            upstream_revision}))
+            elif not native and config.upstream_branch:
+                upstream_branch = Branch.open(config.upstream_branch)
+                upstream_sources.append(UpstreamBranchSource(upstream_branch))
+            upstream_sources.extend([
+                GetOrigSourceSource(tree, larstiq), 
+                UScanSource(tree, larstiq),
+                ])
+            if split:
+                upstream_sources.append(SelfSplitSource(tree))
+ 
+            upstream_provider = UpstreamProvider(changelog.package,
+                changelog.version, orig_dir, upstream_sources)
 
             if merge:
                 distiller_cls = MergeModeDistiller
@@ -432,68 +461,225 @@ class cmd_merge_upstream(Command):
 
     You must supply the source to import from, and the version number of the
     new release. The source can be a .tar.gz, .tar, .tar.bz2, .tgz or .zip
-    archive, or a directory. The source may also be a remote file.
+    archive, or a directory. The source may also be a remote file described by
+    a URL.
 
     You must supply the version number of the new upstream release
-    using --version.
+    using --version, unless you're importing from an upstream branch, in which
+    case it can be guessed from that.
 
     The distribution this version is targetted at can be specified with
-    --distribution. This will be used to guess the version number, so you
-    can always correct it in the changelog.
+    --distribution. This will be used to guess the version number suffix
+    that you want, but you can always correct it in the resulting
+    debian/changelog.
 
     If there is no debian changelog in the branch to retrieve the package
     name from then you must pass the --package option. If this version
     will change the name of the source package then you can use this option
     to set the new name.
+
+    examples::
+
+        bzr merge-upstream --version 0.2 \
+            http://example.org/releases/scruff-0.2.tar.gz
+
+    If you are merging a branch as well as the tarball then you can
+    specify the branch after the tarball, along with -r to specify the
+    revision of that branch to take::
+
+        bzr merge-upstream --version 0.2 \
+            http://example.org/releases/scruff-0.2.tar.gz \
+            http://scruff.org/bzr/scruff.dev -r tag:0.2
+
+    If there is no upstream release tarball, and you want bzr-builddeb to
+    create the tarball for you::
+
+        bzr merge-upstream --version 0.2 http://scruff.org/bzr/scruff.dev
+
+    Note that the created tarball is just the same as the contents of
+    the branch at the specified revision. If you wish to have something
+    different, for instance the results of running "make dist", then you
+    should create the tarball first, and pass it to the command as in
+    the second example.
     """
     takes_args = ['location?', 'upstream_branch?']
     aliases = ['mu']
 
     package_opt = Option('package', help="The name of the source package.",
                          type=str)
-    version_opt = Option('version', help="The version number of this release.",
-                         type=str)
+    version_opt = Option('version',
+        help="The upstream version number of this release, for example "
+        "\"0.2\".", type=str)
     distribution_opt = Option('distribution', help="The distribution that "
             "this release is targetted at.", type=str)
     directory_opt = Option('directory',
                            help='Working tree into which to merge.',
                            short_name='d', type=unicode)
     last_version_opt = Option('last-version',
-                              help='The previous version that was merged..',
-                              type=str)
+                              help='The full version of the last time '
+                              'upstream was merged.', type=str)
     force_opt = Option('force',
                        help=('Force a merge even if the upstream branch '
                              'has not changed.'))
     v3_opt = Option('v3', help='Use dpkg-source format v3.')
 
 
-    takes_options = [package_opt, no_user_conf_opt, version_opt,
+    takes_options = [package_opt, version_opt,
             distribution_opt, directory_opt, last_version_opt,
             force_opt, v3_opt, 'revision', 'merge-type']
 
-    def _update_changelog(self, tree, version, distribution_name, changelog,
-            package):
-        from bzrlib.plugins.builddeb.merge_upstream import package_version
-        if "~bzr" in str(version) or "+bzr" in str(version):
-            entry_description = "New upstream snapshot."
-        else:
-            entry_description = "New upstream release."
-        proc = subprocess.Popen(["dch", "-v",
-                str(package_version(version, distribution_name)),
-                "-D", "UNRELEASED", "--release-heuristic", "changelog",
-                entry_description], cwd=tree.basedir)
-        proc.wait()
-        if proc.returncode != 0:
-            raise BzrCommandError('Adding a new changelog stanza after the '
-                    'merge had completed failed. Add the new changelog entry '
-                    'yourself, review the merge, and then commit.')
 
-    def run(self, location=None, upstream_branch=None, version=None, distribution=None,
-            package=None, no_user_config=None, directory=".", revision=None,
-            merge_type=None, last_version=None, force=None, v3=None):
-        from bzrlib.plugins.builddeb.errors import MissingChangelogError
+    def _add_changelog_entry(self, tree, package, version, distribution_name,
+            changelog):
+        from bzrlib.plugins.builddeb.merge_upstream import (
+            changelog_add_new_version)
+        if not changelog_add_new_version(tree, version, distribution_name,
+                changelog, package):
+            raise BzrCommandError('Adding a new changelog stanza after the '
+                    'merge had completed failed. Add the new changelog '
+                    'entry yourself, review the merge, and then commit.')
+
+    def _do_merge(self, tree, tarball_filename, version, current_version,
+            upstream_branch, upstream_revision, merge_type, force):
+        db = DistributionBranch(tree.branch, None, tree=tree)
+        dbs = DistributionBranchSet()
+        dbs.add_branch(db)
+        conflicts = db.merge_upstream(tarball_filename, version,
+                current_version, upstream_branch=upstream_branch,
+                upstream_revision=upstream_revision,
+                merge_type=merge_type, force=force)
+        return conflicts
+
+    def _export_tarball(self, package, version, orig_dir, upstream_branch,
+            upstream_revision):
+        # TODO: a way to use bz2 on export
+        dest_name = tarball_name(package, version)
+        tarball_filename = os.path.join(orig_dir, dest_name)
+        upstream = UpstreamBranchSource(upstream_branch,
+                {version: upstream_revision})
+        upstream.get_specific_version(package, version, orig_dir)
+        return tarball_filename
+
+    def _fetch_tarball(self, package, version, orig_dir, location, v3):
         from bzrlib.plugins.builddeb.repack_tarball import repack_tarball
-        from bzrlib.plugins.builddeb.merge_upstream import upstream_branch_version
+        format = None
+        if v3:
+            if (location.endswith(".tar.bz2")
+                    or location.endswith(".tbz2")):
+                format = "bz2"
+        dest_name = tarball_name(package, version, format=format)
+        tarball_filename = os.path.join(orig_dir, dest_name)
+        try:
+            repack_tarball(location, dest_name, target_dir=orig_dir,
+                    force_gz=not v3)
+        except FileExists:
+            raise BzrCommandError("The target file %s already exists, and is either "
+                                  "different to the new upstream tarball, or they "
+                                  "are of different formats. Either delete the target "
+                                  "file, or use it as the argument to import."
+                                  % dest_name)
+        return tarball_filename
+
+    def _get_tarball(self, config, tree, package, version, upstream_branch,
+            upstream_revision, no_tarball, v3, location):
+        orig_dir = config.orig_dir or default_orig_dir
+        orig_dir = os.path.join(tree.basedir, orig_dir)
+        if not os.path.exists(orig_dir):
+            os.makedirs(orig_dir)
+        if upstream_branch and no_tarball:
+            tarball_filename = self._export_tarball(package, version,
+                orig_dir, upstream_branch, upstream_revision)
+        else:
+            tarball_filename = self._fetch_tarball(package, version, orig_dir,
+                location, v3)
+        return tarball_filename
+
+    def _get_version(self, version, package, no_tarball, upstream_branch,
+            upstream_revision, current_version):
+        from bzrlib.plugins.builddeb.merge_upstream import (
+            upstream_branch_version)
+        if version is None:
+            if upstream_branch and no_tarball:
+                version = str(upstream_branch_version(upstream_branch,
+                        upstream_revision, package,
+                        current_version))
+                note("Using version string %s for upstream branch." % (version))
+            else:
+                raise BzrCommandError("You must specify the "
+                        "version number using --version.")
+        return version
+
+    def _get_upstream_revision(self, upstream_branch, revision):
+        upstream_revision = None
+        if upstream_branch is not None:
+            if revision is not None:
+                if len(revision) > 1:
+                    raise BzrCommandError("merge-upstream takes only a single --revision")
+                upstream_revspec = revision[0]
+                upstream_revision = upstream_revspec.as_revision_id(upstream_branch)
+            else:
+                upstream_revision = upstream_branch.last_revision()
+        return upstream_revision
+
+    def _get_upstream_branch(self, location, upstream_branch, revision):
+        no_tarball = False
+        if upstream_branch is None:
+            try:
+                upstream_branch = Branch.open(location)
+                no_tarball = True
+            except NotBranchError:
+                upstream_branch = None
+                if revision is not None:
+                    raise BzrCommandError("--revision is not allowed when"
+                            " merging only a tarball")
+        else:
+            upstream_branch = Branch.open(upstream_branch)
+        return no_tarball, upstream_branch
+
+    def _get_changelog_info(self, tree, last_version, package, distribution):
+        from bzrlib.plugins.builddeb.errors import MissingChangelogError
+        changelog = None
+        current_version = last_version
+        try:
+            changelog = find_changelog(tree, False, max_blocks=2)[0]
+            if last_version is None:
+                current_version = changelog.version.upstream_version
+            if package is None:
+                package = changelog.package
+            if distribution is None:
+                distribution = find_last_distribution(changelog)
+                if distribution is not None:
+                    note("Using distribution %s" % distribution)
+        except MissingChangelogError:
+            pass
+        if distribution is None:
+            note("No distribution specified, and no changelog, "
+                    "assuming 'debian'")
+            distribution = "debian"
+        if package is None:
+            raise BzrCommandError("You did not specify --package, and "
+                    "there is no changelog from which to determine the "
+                    "package name, which is needed to know the name to "
+                    "give the .orig.tar.gz. Please specify --package.")
+        distribution = distribution.lower()
+        distribution_name = lookup_distribution(distribution)
+        if distribution_name is None:
+            raise BzrCommandError("Unknown target distribution: %s" \
+                        % distribution)
+        return current_version, package, distribution, distribution_name, changelog
+
+    def _get_upstream_location(self, location, config):
+        if location is None:
+            if config.upstream_branch is not None:
+                location = config.upstream_branch
+            else:
+                raise BzrCommandError("No location specified to merge")
+        return location
+
+    def run(self, location=None, upstream_branch=None, version=None,
+            distribution=None, package=None,
+            directory=".", revision=None, merge_type=None,
+            last_version=None, force=None, v3=None):
         tree, _ = WorkingTree.open_containing(directory)
         tree.lock_write()
         try:
@@ -502,7 +688,7 @@ class cmd_merge_upstream(Command):
                 raise BzrCommandError("There are uncommitted changes in the "
                         "working tree. You must commit before using this "
                         "command.")
-            config = debuild_config(tree, tree, no_user_config)
+            config = debuild_config(tree, tree)
             if config.merge:
                 raise BzrCommandError("Merge upstream in merge mode is not "
                         "yet supported.")
@@ -510,117 +696,24 @@ class cmd_merge_upstream(Command):
                 raise BzrCommandError("Merge upstream in native mode is not "
                         "yet supported.")
 
-            if location is None:
-                if config.upstream_branch is not None:
-                    location = config.upstream_branch
-                else:
-                    raise BzrCommandError("No location specified to merge")
-            changelog = None
-            try:
-                changelog = find_changelog(tree, False, max_blocks=2)[0]
-                if last_version is None:
-                    current_version = changelog.version
-                else:
-                    current_version = Version(last_version)
-                if package is None:
-                    package = changelog.package
-                if distribution is None:
-                    distribution = find_last_distribution(changelog)
-                    if distribution is not None:
-                        note("Using distribution %s" % distribution)
-            except MissingChangelogError:
-                current_version = None
-            if distribution is None:
-                note("No distribution specified, and no changelog, "
-                        "assuming 'debian'")
-                distribution = "debian"
-
-            if package is None:
-                raise BzrCommandError("You did not specify --package, and "
-                        "there is no changelog from which to determine the "
-                        "package name, which is needed to know the name to "
-                        "give the .orig.tar.gz. Please specify --package.")
-
-            no_tarball = False
-            if upstream_branch is None:
-                try:
-                    upstream_branch = Branch.open(location)
-                    no_tarball = True
-                except NotBranchError:
-                    upstream_branch = None
-            else:
-                upstream_branch = Branch.open(upstream_branch)
-
-            distribution = distribution.lower()
-            distribution_name = lookup_distribution(distribution)
-            if distribution_name is None:
-                raise BzrCommandError("Unknown target distribution: %s" \
-                            % distribution)
-
-            upstream_revision = None
-            if upstream_branch:
-                if revision is not None:
-                    if len(revision) > 1:
-                        raise BzrCommandError("merge-upstream takes only a single --revision")
-                    upstream_revspec = revision[0]
-                    upstream_revision = upstream_revspec.as_revision_id(upstream_branch)
-                else:
-                    upstream_revision = upstream_branch.last_revision()
-
-            if no_tarball and revision is not None:
-                raise BzrCommandError("--revision is not allowed when merging a tarball")
-
-            if version is None:
-                if upstream_branch and no_tarball:
-                    version = upstream_branch_version(upstream_branch,
-                            upstream_revision, package,
-                            current_version.upstream_version)
-                    note("Using version string %s for upstream branch." % (version))
-                else:
-                    raise BzrCommandError("You must specify the "
-                            "version number using --version.")
-
-            version = Version(version)
-            orig_dir = config.orig_dir or default_orig_dir
-            orig_dir = os.path.join(tree.basedir, orig_dir)
-            if not os.path.exists(orig_dir):
-                os.makedirs(orig_dir)
-            if upstream_branch and no_tarball:
-                # TODO: a way to use bz2 on export
-                dest_name = tarball_name(package, version.upstream_version)
-                tarball_filename = os.path.join(orig_dir, dest_name)
-                upstream = UpstreamBranchSource(upstream_branch,
-                        upstream_revision)
-                upstream.get_specific_version(package,
-                        version.upstream_version, orig_dir)
-            else:
-                format = None
-                if v3:
-                    if (location.endswith(".tar.bz2")
-                            or location.endswith(".tbz2")):
-                        format = "bz2"
-                dest_name = tarball_name(package, version.upstream_version,
-                        format=format)
-                tarball_filename = os.path.join(orig_dir, dest_name)
-                try:
-                    repack_tarball(location, dest_name, target_dir=orig_dir,
-                            force_gz=not v3)
-                except FileExists:
-                    raise BzrCommandError("The target file %s already exists, and is either "
-                                          "different to the new upstream tarball, or they "
-                                          "are of different formats. Either delete the target "
-                                          "file, or use it as the argument to import."
-                                          % dest_name)
-            db = DistributionBranch(tree.branch, None, tree=tree)
-            dbs = DistributionBranchSet()
-            dbs.add_branch(db)
-            conflicts = db.merge_upstream(tarball_filename, version,
-                    current_version, upstream_branch=upstream_branch,
-                    upstream_revision=upstream_revision,
-                    merge_type=merge_type, force=force)
-
-            self._update_changelog(tree, version, distribution_name, changelog,
-                    package)
+            location = self._get_upstream_location(location, config)
+            (current_version, package, distribution, distribution_name,
+             changelog) = self._get_changelog_info( tree, last_version,
+                 package, distribution)
+            no_tarball, upstream_branch = self._get_upstream_branch(
+                location, upstream_branch, revision)
+            upstream_revision = self._get_upstream_revision(upstream_branch,
+                revision)
+            version = self._get_version(version, package, no_tarball,
+                upstream_branch, upstream_revision, current_version)
+            tarball_filename = self._get_tarball(config, tree, package,
+                version, upstream_branch, upstream_revision, no_tarball, v3,
+                location)
+            conflicts = self._do_merge(tree, tarball_filename, version,
+                current_version, upstream_branch, upstream_revision,
+                merge_type, force)
+            self._add_changelog_entry(tree, package, version,
+                distribution_name, changelog)
         finally:
             tree.unlock()
         note("The new upstream version has been imported.")
@@ -709,7 +802,7 @@ class cmd_import_dsc(Command):
                 raise BzrCommandError("You must give the location of at least one "
                                       "source package to install, or use the "
                                       "--file option.")
-            config = debuild_config(tree, tree, False)
+            config = debuild_config(tree, tree)
             if config.merge:
                 raise BzrCommandError("import-dsc in merge mode is not "
                         "yet supported.")
@@ -729,9 +822,10 @@ class cmd_import_dsc(Command):
                 if last_version is not None:
                     if not db.has_upstream_version_in_packaging_branch(
                             last_version.upstream_version):
-                        raise BzrCommandError("Unable to find the tag for "
-                                "the previous upstream version, %s, in the "
-                                "branch: %s" % (last_version,
+                        raise BzrCommandError("Unable to find the tag for the "
+                            "previous upstream version, %s, in the branch: %s."
+                            " Consider importing it via import-dsc or "
+                            "import-upstream." % (last_version,
                                     db.upstream_tag_name(last_version.upstream_version)))
                     upstream_tip = db.revid_of_upstream_version_from_branch(
                             last_version.upstream_version)
@@ -745,6 +839,95 @@ class cmd_import_dsc(Command):
             tree.unlock()
 
 
+class cmd_import_upstream(Command):
+    """Imports an upstream tarball.
+
+    This will import an upstream tarball in to your branch, but not modify the
+    working tree. Use merge-upstream if you wish to directly merge the new
+    upstream version in to your tree.
+
+    The imported revision can be accessed using the tag name that will be
+    reported at the end of a successful operation. The revision will include
+    the pristine-tar data that will allow other commands to recreate the
+    tarball when needed.
+
+    For instance::
+
+        $ bzr import-upstream 1.2.3 ../package_1.2.3.orig.tar.gz
+
+    If upstream is packaged in bzr, you should provide the upstream branch
+    whose tip commit is the closest match to the tarball::
+
+        $ bzr import-upstream 1.2.3 ../package_1.2.3.orig.tar.gz ../upstream
+
+    After doing this, commands that assume there is an upstream tarball, like
+    'bzr builddeb' will be able to recreate the one provided at import-upstream
+    time, meaning that you don't need to distribute the tarball in addition to
+    the branch.
+
+    If you want to manually merge with the imported upstream, you can do::
+
+        $ bzr merge . -r tag:upstream-1.2.3
+
+    The imported revision will have file ids taken from your branch, the
+    upstream branch, or previous tarball imports as necessary. In addition
+    the parents of the new revision will be the previous upstream tarball
+    import and the tip of the upstream branch if you supply one.
+    """
+
+    takes_options = ['revision']
+
+    takes_args = ['version', 'location', 'upstream_branch?']
+
+    def run(self, version, location, upstream_branch=None, revision=None):
+        # TODO: search for similarity etc.
+        version = version.encode('utf8')
+        branch, _ = Branch.open_containing('.')
+        if upstream_branch is None:
+            upstream = None
+        else:
+            upstream = Branch.open(upstream_branch)
+        branch.lock_write() # we will be adding a tag here.
+        self.add_cleanup(branch.unlock)
+        tempdir = tempfile.mkdtemp(
+            dir=branch.bzrdir.root_transport.clone('..').local_abspath('.'))
+        self.add_cleanup(shutil.rmtree, tempdir)
+        db = DistributionBranch(branch, upstream_branch=upstream)
+        if db.has_upstream_version_in_packaging_branch(version):
+            raise BzrCommandError("Version %s is already present." % version)
+        tagged_versions = {}
+        for tag_name, tag_revid in branch.tags.get_tag_dict().iteritems():
+            if not is_upstream_tag(tag_name):
+                continue
+            tag_version = Version(upstream_tag_version(tag_name))
+            tagged_versions[tag_version] = tag_revid
+        tag_order = sorted(tagged_versions.keys())
+        if tag_order:
+            parents = [tagged_versions[tag_order[-1]]]
+        else:
+            parents = []
+        if parents:
+            # See bug lp:309682
+            upstream.repository.fetch(branch.repository, parents[0])
+            db.extract_upstream_tree(parents[0], tempdir)
+        else:
+            db._create_empty_upstream_tree(tempdir)
+        tree = db.get_branch_tip_revtree()
+        tree.lock_read()
+        dbs = DistributionBranchSet()
+        dbs.add_branch(db)
+        if revision is None:
+            upstream_revid = None
+        elif len(revision) == 1:
+            upstream_revid = revision[0].in_history(upstream).rev_id
+        else:
+            raise BzrCommandError('bzr import-upstream --revision takes exactly'
+                                  ' one revision specifier.')
+        tag_name, _ = db.import_upstream_tarball(location, version, parents,
+            upstream_branch=upstream, upstream_revision=upstream_revid)
+        self.outf.write('Imported %s as tag:%s.\n' % (location, tag_name))
+
+
 class cmd_bd_do(Command):
     """Run a command in an exported package, copying the result back.
 
@@ -774,7 +957,7 @@ class cmd_bd_do(Command):
 
     def run(self, command_list=None):
         t = WorkingTree.open_containing('.')[0]
-        config = debuild_config(t, t, False)
+        config = debuild_config(t, t)
 
         if not config.merge:
             raise BzrCommandError("This command only works for merge mode "
@@ -795,9 +978,13 @@ class cmd_bd_do(Command):
         orig_dir = config.orig_dir
         if orig_dir is None:
             orig_dir = default_orig_dir
+
         upstream_provider = UpstreamProvider(changelog.package, 
                 changelog.version.upstream_version, orig_dir, 
-                get_upstream_sources(t, t.branch, larstiq=larstiq))
+                [PristineTarSource(t, t.branch),
+                 AptSource(),
+                 GetOrigSourceSource(t, larstiq),
+                 UScanSource(t, larstiq) ])
 
         distiller = MergeModeDistiller(t, upstream_provider,
                 larstiq=larstiq)
@@ -841,7 +1028,7 @@ class cmd_bd_do(Command):
 
 class cmd_mark_uploaded(Command):
     """Mark that this branch has been uploaded, prior to pushing it.
-    
+
     When a package has been uploaded we want to mark the revision
     that it was uploaded in. This command automates doing that
     by marking the current tip revision with the version indicated
@@ -850,9 +1037,9 @@ class cmd_mark_uploaded(Command):
     force = Option('force', help="Mark the upload even if it is already "
             "marked.")
 
-    takes_options = [merge_opt, no_user_conf_opt, force]
+    takes_options = [merge_opt, force]
 
-    def run(self, merge=False, no_user_config=False, force=None):
+    def run(self, merge=False, force=None):
         t = WorkingTree.open_containing('.')[0]
         t.lock_write()
         try:
@@ -860,7 +1047,7 @@ class cmd_mark_uploaded(Command):
               raise BzrCommandError("There are uncommitted changes in the "
                       "working tree. You must commit before using this "
                       "command")
-            config = debuild_config(t, t, no_user_config)
+            config = debuild_config(t, t)
             if not merge:
                 merge = config.merge
             (changelog, larstiq) = find_changelog(t, merge)
@@ -896,11 +1083,10 @@ class cmd_merge_package(Command):
     takes_args = ['source']
 
     def run(self, source):
-        source_branch = target_branch = None
+        source_branch = None
         # Get the target branch.
         try:
             tree = WorkingTree.open_containing('.')[0]
-            target_branch = tree.branch
         except (NotBranchError, NoWorkingTree):
             raise BzrCommandError(
                 "There is no tree to merge the source branch in to")
@@ -911,28 +1097,24 @@ class cmd_merge_package(Command):
             raise BzrCommandError("Invalid source branch URL?")
 
         tree.lock_write()
-        try:
-            this_config = debuild_config(tree, tree, False)
-            source_branch.lock_read()
-            try:
-                that_config = debuild_config(source_branch.basis_tree(),
-                        source_branch.basis_tree(), False)
-                if not (this_config.native or that_config.native):
-                    fix_ancestry_as_needed(tree, source_branch)
+        self.add_cleanup(tree.unlock)
+        source_branch.lock_read()
+        self.add_cleanup(source_branch.unlock)
+        this_config = debuild_config(tree, tree)
+        that_config = debuild_config(source_branch.basis_tree(),
+                source_branch.basis_tree())
+        if not (this_config.native or that_config.native):
+            fix_ancestry_as_needed(tree, source_branch)
 
-                # Merge source packaging branch in to the target packaging branch.
-                _merge_tags_if_possible(source_branch, tree.branch)
-                conflicts = tree.merge_from_branch(source_branch)
-                if conflicts > 0:
-                    note('The merge resulted in %s conflicts. Please resolve these '
-                         'and commit the changes with "bzr commit".' % conflicts)
-                else:
-                    note('The merge resulted in no conflicts. You may commit the '
-                    'changes by running "bzr commit".')
-            finally:
-                source_branch.unlock()
-        finally:
-            tree.unlock()
+        # Merge source packaging branch in to the target packaging branch.
+        _merge_tags_if_possible(source_branch, tree.branch)
+        conflicts = tree.merge_from_branch(source_branch)
+        if conflicts > 0:
+            note('The merge resulted in %s conflicts. Please resolve these '
+                 'and commit the changes with "bzr commit".' % conflicts)
+        else:
+            note('The merge resulted in no conflicts. You may commit the '
+            'changes by running "bzr commit".')
 
 
 class cmd_dh_make(Command):
