@@ -30,10 +30,6 @@ from base64 import (
         standard_b64encode,
         )
 import errno
-try:
-    import hashlib as md5
-except ImportError:
-    import md5
 import os
 import re
 import shutil
@@ -41,8 +37,13 @@ import stat
 import subprocess
 import tempfile
 
-from debian_bundle import deb822
-from debian_bundle.changelog import Version, Changelog, VersionError
+try:
+    from debian import deb822
+    from debian.changelog import Version, Changelog, VersionError
+except ImportError:
+  # Prior to 0.1.15 the debian module was called debian_bundle
+    from debian_bundle import deb822
+    from debian_bundle.changelog import Version, Changelog, VersionError
 
 from bzrlib import (
                     bzrdir,
@@ -53,13 +54,13 @@ from bzrlib.errors import (
         AlreadyBranchError,
         BzrCommandError,
         NotBranchError,
+        NoSuchRevision,
         NoWorkingTree,
         UnrelatedBranches,
         )
-from bzrlib.export import export
 from bzrlib.revisionspec import RevisionSpec
 from bzrlib.revision import NULL_REVISION
-from bzrlib.trace import warning, mutter, note
+from bzrlib.trace import warning, mutter
 from bzrlib.transport import (
     get_transport,
     )
@@ -72,8 +73,10 @@ from bzrlib.plugins.builddeb.errors import (
                 UpstreamBranchAlreadyMerged,
                 )
 from bzrlib.plugins.builddeb.util import (
+    export,
     get_commit_info_from_changelog,
     get_snapshot_revision,
+    md5sum_filename,
     open_file_via_transport,
     open_transport,
     safe_decode,
@@ -125,6 +128,57 @@ class DscComp(object):
     return -1
 
 
+
+def reconstruct_pristine_tar(dest, delta, dest_filename):
+    """Reconstruct a pristine tarball from a directory and a delta.
+
+    :param dest: Directory to pack
+    :param delta: pristine-tar delta
+    :param dest_filename: Destination filename
+    """
+    command = ["pristine-tar", "gentar", "-",
+               os.path.abspath(dest_filename)]
+    try:
+        proc = subprocess.Popen(command, stdin=subprocess.PIPE,
+                cwd=dest, preexec_fn=subprocess_setup,
+                stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
+    except OSError, e:
+        if e.errno == errno.ENOENT:
+            raise PristineTarError("pristine-tar is not installed")
+        else:
+            raise
+    (stdout, stderr) = proc.communicate(delta)
+    if proc.returncode != 0:
+        raise PristineTarError("Generating tar from delta failed: %s" % stdout)
+
+
+def make_pristine_tar_delta(dest, tarball_path):
+    """Create a pristine-tar delta for a tarball.
+
+    :param dest: Directory to generate pristine tar delta for
+    :param tarball_path: Path to the tarball
+    :return: pristine-tarball
+    """
+    # If tarball_path is relative, the cwd=dest parameter to Popen will make
+    # pristine-tar faaaail. pristine-tar doesn't use the VFS either, so we
+    # assume local paths.
+    tarball_path = osutils.abspath(tarball_path)
+    command = ["pristine-tar", "gendelta", tarball_path, "-"]
+    try:
+        proc = subprocess.Popen(command, stdout=subprocess.PIPE,
+                cwd=dest, preexec_fn=subprocess_setup,
+                stderr=subprocess.PIPE)
+    except OSError, e:
+        if e.errno == errno.ENOENT:
+            raise PristineTarError("pristine-tar is not installed")
+        else:
+            raise
+    (stdout, stderr) = proc.communicate()
+    if proc.returncode != 0:
+        raise PristineTarError("Generating delta from tar failed: %s" % stderr)
+    return stdout
+
+
 class DistributionBranchSet(object):
     """A collection of DistributionBranches with an ordering.
 
@@ -343,28 +397,16 @@ class DistributionBranch(object):
         :return: True if the upstream branch contains the specified upstream
             version of the package. False otherwise.
         """
-        tag_name = self.upstream_tag_name(version)
-        if self._has_version(self.upstream_branch, tag_name, md5=md5):
-            return True
-        tag_name = self.upstream_tag_name(version, distro="debian")
-        if self._has_version(self.upstream_branch, tag_name, md5=md5):
-            return True
-        tag_name = self.upstream_tag_name(version, distro="ubuntu")
-        if self._has_version(self.upstream_branch, tag_name, md5=md5):
-            return True
+        for tag_name in self.possible_upstream_tag_names(version):
+            if self._has_version(self.upstream_branch, tag_name, md5=md5):
+                return True
         return False
 
     def has_upstream_version_in_packaging_branch(self, version, md5=None):
-        assert isinstance(version, str)
-        tag_name = self.upstream_tag_name(version)
-        if self._has_version(self.branch, tag_name, md5=md5):
-            return True
-        tag_name = self.upstream_tag_name(version, distro="debian")
-        if self._has_version(self.branch, tag_name, md5=md5):
-            return True
-        tag_name = self.upstream_tag_name(version, distro="ubuntu")
-        if self._has_version(self.branch, tag_name, md5=md5):
-            return True
+        assert isinstance(version, str), str(type(version))
+        for tag_name in self.possible_upstream_tag_names(version):
+            if self._has_version(self.branch, tag_name, md5=md5):
+                return True
         return False
 
     def contained_versions(self, versions):
@@ -463,18 +505,19 @@ class DistributionBranch(object):
         :return: the revision id corresponding to the upstream portion
             of the version
         """
-        tag_name = self.upstream_tag_name(version)
-        if self._has_version(self.upstream_branch, tag_name):
-            return self.upstream_branch.tags.lookup_tag(tag_name)
-        tag_name = self.upstream_tag_name(version, distro="debian")
-        if self._has_version(self.upstream_branch, tag_name):
-            return self.upstream_branch.tags.lookup_tag(tag_name)
-        tag_name = self.upstream_tag_name(version, distro="ubuntu")
-        if self._has_version(self.upstream_branch, tag_name):
-            return self.upstream_branch.tags.lookup_tag(tag_name)
+        for tag_name in self.possible_upstream_tag_names(version):
+            if self._has_version(self.upstream_branch, tag_name):
+                return self.upstream_branch.tags.lookup_tag(tag_name)
         tag_name = self.upstream_tag_name(version)
         return self.upstream_branch.tags.lookup_tag(tag_name)
 
+    def possible_upstream_tag_names(self, version):
+        tags = [self.upstream_tag_name(version),
+                self.upstream_tag_name(version, distro="debian"),
+                self.upstream_tag_name(version, distro="ubuntu"),
+                "upstream/%s" % self.tag_name(version)]
+        return tags
+
     def tag_version(self, version, revid=None):
         """Tags the branch's last revision with the given version.
 
@@ -495,20 +538,29 @@ class DistributionBranch(object):
     def tag_upstream_version(self, version, revid=None):
         """Tags the upstream branch's last revision with an upstream version.
 
-        Sets a tag on the last revision of the upstream branch with a tag
-        that refers to the upstream part of the version provided.
+        Sets a tag on the last revision of the upstream branch and on the main
+        branch with a tag that refers to the upstream part of the version
+        provided.
 
         :param version: the upstream part of the version number to derive the 
             tag name from.
         :param revid: the revid to associate the tag with, or None for the
             tip of self.upstream_branch.
+        :return The tag name, revid of the added tag.
         """
         assert isinstance(version, str)
         tag_name = self.upstream_tag_name(version)
         if revid is None:
             revid = self.upstream_branch.last_revision()
         self.upstream_branch.tags.set_tag(tag_name, revid)
+        try:
+            self.branch.repository.fetch(self.upstream_branch.repository,
+                revision_id=revid)
+        except NoSuchRevision:
+            # See bug lp:574223
+            pass
         self.branch.tags.set_tag(tag_name, revid)
+        return tag_name, revid
 
     def _default_config_for_tree(self, tree):
         # FIXME: shouldn't go to configobj directly
@@ -527,7 +579,6 @@ class DistributionBranch(object):
                 tree.unlock()
         return config
 
-
     def _is_tree_native(self, tree):
         config = self._default_config_for_tree(tree)
         if config is not None:
@@ -674,7 +725,7 @@ class DistributionBranch(object):
         """Return the list of parents for a specific version.
 
         This method returns the list of revision ids that should be parents
-        for importing a specifc package version. The specific package version
+        for importing a specific package version. The specific package version
         is the first element of the list of versions passed.
 
         The parents are determined by looking at the other versions in the
@@ -878,6 +929,7 @@ class DistributionBranch(object):
         :param upstream_parents: the parents to give the upstream revision
         :param timestamp: a tuple of (timestamp, timezone) to use for
             the commit, or None to use the current time.
+        :return: (tag_name, revision_id) of the imported tarball.
         """
         # Should we just dump the upstream part on whatever is currently
         # there, or try and pull all of the other upstream versions
@@ -900,6 +952,7 @@ class DistributionBranch(object):
             return br.repository.revision_tree(br.last_revision())
         upstream_trees = [get_last_revision_tree(o.upstream_branch)
             for o in other_branches]
+        target_tree = None
         if upstream_branch is not None:
             if upstream_revision is None:
                 upstream_revision = upstream_branch.last_revision()
@@ -907,13 +960,22 @@ class DistributionBranch(object):
                     last_revision=upstream_revision)
             upstream_branch.tags.merge_to(self.upstream_branch.tags)
             upstream_parents.append(upstream_revision)
-            upstream_trees.insert(0,
-                    self.upstream_branch.repository.revision_tree(
-                        upstream_revision))
+            target_tree = self.upstream_branch.repository.revision_tree(
+                        upstream_revision)
         if file_ids_from is not None:
             upstream_trees = file_ids_from + upstream_trees
-        import_dir(self.upstream_tree, upstream_part,
-                file_ids_from=upstream_trees + [self.tree])
+        if self.tree:
+            self_tree = self.tree
+            self_tree.lock_write() # might also be upstream tree for dh_make
+        else:
+            self_tree = self.get_branch_tip_revtree()
+            self_tree.lock_read()
+        try:
+            import_dir(self.upstream_tree, upstream_part,
+                    file_ids_from=[self_tree] + upstream_trees,
+                    target_tree=target_tree)
+        finally:
+            self_tree.unlock()
         self.upstream_tree.set_parent_ids(upstream_parents)
         revprops = {"deb-md5": md5}
         if upstream_tarball is not None:
@@ -933,12 +995,41 @@ class DistributionBranch(object):
         revid = self.upstream_tree.commit("Import upstream version %s" \
                 % (version,),
                 revprops=revprops, timestamp=timestamp, timezone=timezone)
-        self.tag_upstream_version(version, revid=revid)
-        return revid
+        tag_name, _ = self.tag_upstream_version(version, revid=revid)
+        return tag_name, revid
+
+    def import_upstream_tarball(self, tarball_filename, version, parents,
+        md5sum=None, upstream_branch=None, upstream_revision=None):
+        """Import an upstream part to the upstream branch.
+
+        :param tarball_filename: The tarball to import.
+        :param version: The upstream version to import.
+        :param parents: The tarball-branch parents to use for the import.
+            If an upstream branch is supplied, its automatically added to
+            parents.
+        :param upstream_branch: An upstream branch to associate with the
+            tarball.
+        :param upstream_revision: Upstream revision id
+        :param md5sum: hex digest of the md5sum of the tarball, if known.
+        :return: (tag_name, revision_id) of the imported tarball.
+        """
+        if not md5sum:
+            md5sum = md5sum_filename(tarball_filename)
+        tarball_dir = self._extract_tarball_to_tempdir(tarball_filename)
+        try:
+            return self.import_upstream(tarball_dir, version, md5sum, parents,
+                upstream_tarball=tarball_filename,
+                upstream_branch=upstream_branch,
+                upstream_revision=upstream_revision)
+        finally:
+            shutil.rmtree(tarball_dir)
+
+    def get_branch_tip_revtree(self):
+        return self.branch.repository.revision_tree(
+                self.branch.last_revision())
 
     def _mark_native_config(self, native):
-        poss_native_tree = self.branch.repository.revision_tree(
-                self.branch.last_revision())
+        poss_native_tree = self.get_branch_tip_revtree()
         current_native = self._is_tree_native(poss_native_tree)
         current_config = self._default_config_for_tree(poss_native_tree)
         dirname = os.path.join(self.tree.basedir,
@@ -1152,7 +1243,7 @@ class DistributionBranch(object):
                     # from another branch:
                     upstream_parents = self.upstream_parents(versions,
                             version.upstream_version)
-                    new_revid = self.import_upstream(upstream_part,
+                    _, new_revid = self.import_upstream(upstream_part,
                             version.upstream_version,
                             upstream_md5, upstream_parents,
                             upstream_tarball=upstream_tarball,
@@ -1284,13 +1375,14 @@ class DistributionBranch(object):
             extractor.cleanup()
 
     def extract_upstream_tree(self, upstream_tip, basedir):
-        # Extract that to a tempdir so we can get a working
-        # tree for it.
+        """Extract upstream_tip to a tempdir as a working tree."""
         # TODO: should stack rather than trying to use the repository,
         # as that will be more efficient.
         # TODO: remove the _extract_upstream_tree alias below.
         to_location = os.path.join(basedir, "upstream")
-        dir_to = self.branch.bzrdir.sprout(to_location,
+        # Use upstream_branch if it has been set, otherwise self.branch.
+        source_branch = self.upstream_branch or self.branch
+        dir_to = source_branch.bzrdir.sprout(to_location,
                 revision_id=upstream_tip,
                 accelerator_tree=self.tree)
         try:
@@ -1324,14 +1416,24 @@ class DistributionBranch(object):
                 existing_bzrdir.create_workingtree()
         self.upstream_branch = branch
         self.upstream_tree = branch.bzrdir.open_workingtree()
-        self.upstream_tree.set_root_id(self.tree.path2id(""))
+        if self.tree:
+            root_id = self.tree.path2id('')
+        else:
+            tip = self.get_branch_tip_revtree()
+            tip.lock_read()
+            try:
+                root_id = tip.path2id('')
+            finally:
+                tip.unlock()
+        if root_id:
+            self.upstream_tree.set_root_id(root_id)
 
     def _extract_tarball_to_tempdir(self, tarball_filename):
         tempdir = tempfile.mkdtemp()
         if tarball_filename.endswith(".tar.bz2"):
-	    tar_args = 'xjf'
+            tar_args = 'xjf'
         else:
-	    tar_args = 'xzf'
+            tar_args = 'xzf'
         try:
             proc = subprocess.Popen(["tar", tar_args, tarball_filename, "-C",
                     tempdir, "--strip-components", "1"],
@@ -1351,54 +1453,58 @@ class DistributionBranch(object):
     def revid_of_upstream_version_from_branch(self, version):
         # TODO: remove the _revid_of_upstream_version_from_branch alias below.
         assert isinstance(version, str)
-        tag_name = self.upstream_tag_name(version)
-        if self._has_version(self.branch, tag_name):
-            return self.branch.tags.lookup_tag(tag_name)
-        tag_name = self.upstream_tag_name(version, distro="debian")
-        if self._has_version(self.branch, tag_name):
-            return self.branch.tags.lookup_tag(tag_name)
-        tag_name = self.upstream_tag_name(version, distro="ubuntu")
-        if self._has_version(self.branch, tag_name):
-            return self.branch.tags.lookup_tag(tag_name)
+        for tag_name in self.possible_upstream_tag_names(version):
+            if self._has_version(self.branch, tag_name):
+                return self.branch.tags.lookup_tag(tag_name)
         tag_name = self.upstream_tag_name(version)
         return self.branch.tags.lookup_tag(tag_name)
 
     _revid_of_upstream_version_from_branch = revid_of_upstream_version_from_branch
 
+    def _export_previous_upstream_tree(self, previous_version, tempdir, upstream_branch=None):
+        assert isinstance(previous_version, str), \
+            "Should pass upstream version as str, not Version."
+        previous_upstream_revision = get_snapshot_revision(previous_version)
+        if self.has_upstream_version_in_packaging_branch(previous_version):
+            upstream_tip = self.revid_of_upstream_version_from_branch(
+                    previous_version)
+            self.extract_upstream_tree(upstream_tip, tempdir)
+        elif (upstream_branch is not None and
+              previous_upstream_revision is not None):
+            upstream_tip = RevisionSpec.from_string(previous_upstream_revision).as_revision_id(upstream_branch)
+            assert isinstance(upstream_tip, str)
+            self.extract_upstream_tree(upstream_tip, tempdir)
+        else:
+            raise BzrCommandError("Unable to find the tag for the "
+                    "previous upstream version, %s, in the branch: "
+                    "%s" % (
+                previous_version,
+                self.upstream_tag_name(previous_version)))
+
     def merge_upstream(self, tarball_filename, version, previous_version,
             upstream_branch=None, upstream_revision=None, merge_type=None,
             force=False):
         assert self.upstream_branch is None, \
                 "Should use self.upstream_branch if set"
+        assert isinstance(version, str), \
+            "Should pass version as str not %s" % str(type(version))
+        assert isinstance(previous_version, str), \
+            "Should pass previous_version as str not %s" % str(
+                    type(previous_version))
         tempdir = tempfile.mkdtemp(dir=os.path.join(self.tree.basedir, '..'))
         try:
             if previous_version is not None:
-                previous_upstream_revision = get_snapshot_revision(previous_version.upstream_version)
-                if self.has_upstream_version_in_packaging_branch(
-                        previous_version.upstream_version):
-                    upstream_tip = self.revid_of_upstream_version_from_branch(
-                            previous_version.upstream_version)
-                    self.extract_upstream_tree(upstream_tip, tempdir)
-                elif (upstream_branch is not None and 
-                      previous_upstream_revision is not None):
-                    upstream_tip = RevisionSpec.from_string(previous_upstream_revision).as_revision_id(upstream_branch)
-                    assert isinstance(upstream_tip, str)
-                    self.extract_upstream_tree(upstream_tip, tempdir)
-                else:
-                    raise BzrCommandError("Unable to find the tag for the "
-                            "previous upstream version, %s, in the branch: "
-                            "%s" % (
-                        previous_version.upstream_version,
-                        self.upstream_tag_name(
-                            previous_version.upstream_version)))
+                self._export_previous_upstream_tree(previous_version, tempdir,
+                    upstream_branch)
             else:
                 self._create_empty_upstream_tree(tempdir)
-            if self.has_upstream_version_in_packaging_branch(version.upstream_version):
+            if self.has_upstream_version_in_packaging_branch(version):
                 raise UpstreamAlreadyImported(version)
+            if upstream_branch is not None:
+                upstream_branch.lock_read()
             try:
                 if upstream_branch is not None:
-                    upstream_branch.lock_read()
-                    if upstream_revision is not None:
+                    if upstream_revision is None:
                         upstream_revision = upstream_branch.last_revision()
                     graph = self.branch.repository.get_graph(
                             other_repository=upstream_branch.repository)
@@ -1406,17 +1512,15 @@ class DistributionBranch(object):
                             self.branch.last_revision()):
                         raise UpstreamBranchAlreadyMerged
                 tarball_filename = os.path.abspath(tarball_filename)
-                m = md5.md5()
-                m.update(open(tarball_filename).read())
-                md5sum = m.hexdigest()
+                md5sum = md5sum_filename(tarball_filename)
                 tarball_dir = self._extract_tarball_to_tempdir(tarball_filename)
                 try:
                     # FIXME: should use upstream_parents()?
                     parents = []
                     if self.upstream_branch.last_revision() != NULL_REVISION:
                         parents = [self.upstream_branch.last_revision()]
-                    new_revid = self.import_upstream(tarball_dir,
-                            version.upstream_version,
+                    _, new_revid = self.import_upstream(tarball_dir,
+                            version,
                             md5sum, parents, upstream_tarball=tarball_filename,
                             upstream_branch=upstream_branch,
                             upstream_revision=upstream_revision)
@@ -1451,29 +1555,26 @@ class DistributionBranch(object):
         finally:
             shutil.rmtree(tempdir)
 
-    def has_pristine_tar_delta(self, revid):
-        rev = self.branch.repository.get_revision(revid)
+    def has_pristine_tar_delta(self, rev):
         return ('deb-pristine-delta' in rev.properties
                 or 'deb-pristine-delta-bz2' in rev.properties)
 
-    def pristine_tar_format(self, revid):
-        rev = self.branch.repository.get_revision(revid)
+    def pristine_tar_format(self, rev):
         if 'deb-pristine-delta' in rev.properties:
             return 'gz'
         elif 'deb-pristine-delta-bz2' in rev.properties:
             return 'bz2'
-        assert self.has_pristine_tar_delta(revid)
+        assert self.has_pristine_tar_delta(rev)
         raise AssertionError("Not handled new delta type in "
                 "pristine_tar_format")
 
-    def pristine_tar_delta(self, revid):
-        rev = self.branch.repository.get_revision(revid)
+    def pristine_tar_delta(self, rev):
         if 'deb-pristine-delta' in rev.properties:
             uuencoded = rev.properties['deb-pristine-delta']
         elif 'deb-pristine-delta-bz2' in rev.properties:
             uuencoded = rev.properties['deb-pristine-delta-bz2']
         else:
-            assert self.has_pristine_tar_delta(revid)
+            assert self.has_pristine_tar_delta(rev)
             raise AssertionError("Not handled new delta type in "
                     "pristine_tar_delta")
         delta = standard_b64decode(uuencoded)
@@ -1486,22 +1587,13 @@ class DistributionBranch(object):
         tmpdir = tempfile.mkdtemp(prefix="builddeb-pristine-")
         try:
             dest = os.path.join(tmpdir, "orig")
-            export(tree, dest, format='dir')
-            delta = self.pristine_tar_delta(revid)
-            command = ["pristine-tar", "gentar", "-",
-                       os.path.abspath(dest_filename)]
-            try:
-                proc = subprocess.Popen(command, stdin=subprocess.PIPE,
-                        cwd=dest, preexec_fn=subprocess_setup,
-                        stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
-            except OSError, e:
-                if e.errno == errno.ENOENT:
-                    raise PristineTarError("pristine-tar is not installed")
-                else:
-                    raise
-            (stdout, stderr) = proc.communicate(delta)
-            if proc.returncode != 0:
-                raise PristineTarError("Generating tar from delta failed: %s" % stdout)
+            rev = self.branch.repository.get_revision(revid)
+            if self.has_pristine_tar_delta(rev):
+                export(tree, dest, format='dir')
+                delta = self.pristine_tar_delta(rev)
+                reconstruct_pristine_tar(dest, delta, dest_filename)
+            else:
+                export(tree, dest_filename, require_per_file_timestamps=True)
         finally:
             shutil.rmtree(tmpdir)
 
@@ -1516,20 +1608,7 @@ class DistributionBranch(object):
                 export(tree, dest, format='dir')
             finally:
                 tree.unlock()
-            command = ["pristine-tar", "gendelta", tarball_path, "-"]
-            try:
-                proc = subprocess.Popen(command, stdout=subprocess.PIPE,
-                        cwd=dest, preexec_fn=subprocess_setup,
-                        stderr=subprocess.PIPE)
-            except OSError, e:
-                if e.errno == errno.ENOENT:
-                    raise PristineTarError("pristine-tar is not installed")
-                else:
-                    raise
-            (stdout, stderr) = proc.communicate()
-            if proc.returncode != 0:
-                raise PristineTarError("Generating delta from tar failed: %s" % stderr)
-            return stdout
+            return make_pristine_tar_delta(dest, tarball_path)
         finally:
             shutil.rmtree(tmpdir)
 
