@@ -28,7 +28,11 @@ import doctest
 import os
 from unittest import TestSuite
 
-from debian_bundle.changelog import Version, Changelog
+try:
+    from debian.changelog import Version, Changelog
+except ImportError:
+    # Prior to 0.1.15 the debian module was called debian_bundle
+    from debian_bundle.changelog import Version, Changelog
 
 from bzrlib.tests import TestUtil, multiply_tests, TestCaseWithTransport
 
@@ -112,6 +116,7 @@ def load_tests(standard_tests, module, loader):
     testmod_names = [
             'blackbox',
             'test_builder',
+            'test_bzrtools_import',
             'test_commit_message',
             'test_config',
             'test_dh_make',
@@ -125,6 +130,7 @@ def load_tests(standard_tests, module, loader):
             'test_source_distiller',
             'test_upstream',
             'test_util',
+            'test_tagging',
             ]
     suite.addTest(loader.loadTestsFromModuleNames(["%s.%s" % (__name__, i)
                                             for i in testmod_names]))
@@ -323,6 +329,10 @@ class SourcePackageBuilder(object):
     def basedir(self):
         return self.name + "-" + str(self._cl.version.upstream_version)
 
+    def write_debian_files(self, basedir):
+        self._make_files(self.debian_files, basedir)
+        self._make_files({"debian/changelog": str(self._cl)}, basedir)
+
     def _make_base(self):
         basedir = self.basedir()
         os.mkdir(basedir)
@@ -360,8 +370,7 @@ class SourcePackageBuilder(object):
             else:
                 cmd = ["dpkg-source", "--format=3.0 (native)", "-b",
                         basedir]
-        self._make_files(self.debian_files, basedir)
-        self._make_files({"debian/changelog": str(self._cl)}, basedir)
+        self.write_debian_files(basedir)
         proc = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                 stderr=subprocess.STDOUT)
         ret = proc.wait()
