@@ -23,8 +23,13 @@ import os
 import shutil
 import tarfile
 
-from debian_bundle.changelog import Version
-from debian_bundle import deb822
+try:
+    from debian.changelog import Version
+    from debian import deb822
+except ImportError:
+    # Prior to 0.1.15 the debian module was called debian_bundle
+    from debian_bundle.changelog import Version
+    from debian_bundle import deb822
 
 from bzrlib import (
   tests,
@@ -746,15 +751,28 @@ class DistributionBranchTests(BuilddebTestCase):
         check_one_type("modified", modified, changes.modified)
         check_one_type("renamed", renamed, changes.renamed)
 
-    def test_import_upstream(self):
+    def import_a_tree(self, contents=None):
+        """Import a tree from disk."""
         version = Version("0.1-1")
         name = "package"
         basedir = name + "-" + str(version.upstream_version)
-        os.mkdir(basedir)
-        write_to_file(os.path.join(basedir, "README"), "Hi\n")
-        write_to_file(os.path.join(basedir, "BUGS"), "")
+        if contents is None:
+            contents = [
+                (basedir + '/',),
+                (os.path.join(basedir, "README"), "Hi\n"),
+                (os.path.join(basedir, "BUGS"), ""),
+                ]
+        else:
+            # add basedir to the contents
+            contents = [(basedir + '/' + element[0],) + element[1:] for
+                element in contents]
+        self.build_tree_contents(contents)
         self.db1.import_upstream(basedir, version.upstream_version, 
             self.fake_md5_1, [])
+        return version
+
+    def test_import_upstream(self):
+        version = self.import_a_tree()
         tree = self.up_tree1
         branch = tree.branch
         rh = branch.revision_history()
@@ -766,6 +784,15 @@ class DistributionBranchTests(BuilddebTestCase):
                 "Import upstream version %s" % str(version.upstream_version))
         self.assertEqual(rev.properties['deb-md5'], self.fake_md5_1)
 
+    def test_import_upstream_preserves_dot_bzrignore(self):
+        self.import_a_tree([('',), ('.bzrignore', '')])
+        branch = self.up_tree1.branch
+        branch.lock_read()
+        self.addCleanup(branch.unlock)
+        tip = branch.last_revision()
+        revtree = branch.repository.revision_tree(tip)
+        self.assertNotEqual(None, revtree.path2id('.bzrignore'))
+
     def test_import_upstream_on_another(self):
         version1 = Version("0.1-1")
         version2 = Version("0.2-1")
@@ -1472,8 +1499,7 @@ class DistributionBranchTests(BuilddebTestCase):
             tf.add("a")
         finally:
             tf.close()
-        conflicts = db.merge_upstream(tarball_filename, Version("0.2"),
-                Version("0.1"))
+        conflicts = db.merge_upstream(tarball_filename, "0.2", "0.1")
         self.assertEqual(0,  conflicts)
 
     def test_merge_upstream_initial_with_branch(self):
@@ -1510,8 +1536,8 @@ class DistributionBranchTests(BuilddebTestCase):
             tf.add("a")
         finally:
             tf.close()
-        conflicts = db.merge_upstream(tarball_filename, Version("0.2"),
-                Version("0.1"), upstream_branch=upstream_tree.branch,
+        conflicts = db.merge_upstream(tarball_filename, "0.2", "0.1",
+                upstream_branch=upstream_tree.branch,
                 upstream_revision=upstream_rev)
         self.assertEqual(0,  conflicts)
 
@@ -1550,12 +1576,166 @@ class DistributionBranchTests(BuilddebTestCase):
             tf.add("a")
         finally:
             tf.close()
-        conflicts = db.merge_upstream(tarball_filename, Version("0.2"),
-                Version("0.1"), upstream_branch=upstream_tree.branch,
+        conflicts = db.merge_upstream(tarball_filename, "0.2", "0.1",
+                upstream_branch=upstream_tree.branch,
                 upstream_revision=upstream_rev)
         # ./debian conflicts.
         self.assertEqual(3,  conflicts)
 
+    def test_merge_upstream_with_unrelated_branch(self):
+        """Check that we can merge-upstream with an unrelated branch.
+
+        We should do this by changing all the file ids to be the same
+        as in the upstream branch, which gives a discontinuity, but
+        makes for a better experience in the future.
+        """
+        self.requireFeature(PristineTarFeature)
+        version1 = Version("1.0-1")
+        version2 = Version("1.1-1")
+        builder = SourcePackageBuilder("package", version1)
+        builder.add_default_control()
+        builder.add_upstream_file("a", "Original a")
+        builder.build()
+        tree = self.make_branch_and_tree(".")
+        packaging_upstream_tree = self.make_branch_and_tree(
+            "packaging-upstream")
+        db = DistributionBranch(tree.branch, packaging_upstream_tree.branch,
+            tree=tree, upstream_tree=packaging_upstream_tree)
+        dbs = DistributionBranchSet()
+        dbs.add_branch(db)
+        db.import_package(builder.dsc_name())
+        builder.new_version(version2)
+        builder.add_upstream_file("a", "New a")
+        builder.build()
+        upstream_tree = self.make_branch_and_tree("upstream")
+        self.build_tree(['upstream/a'])
+        upstream_tree.add(['a'], ['a-id'])
+        upstream_tree.commit("one")
+        upstream_rev = upstream_tree.branch.last_revision()
+        db = DistributionBranch(tree.branch, None, tree=tree)
+        dbs = DistributionBranchSet()
+        dbs.add_branch(db)
+        tree.lock_write()
+        self.addCleanup(tree.unlock)
+        db.merge_upstream(builder.tar_name(), str(version2),
+            version1.upstream_version,
+            upstream_branch=upstream_tree.branch,
+            upstream_revision=upstream_rev)
+        rh1 = tree.branch.revision_history()
+        self.assertEqual(2, len(rh1))
+        packaging_upstream_tip = tree.get_parent_ids()[1]
+        # We added the extra parent for the upstream branch
+        revtree = tree.branch.repository.revision_tree(packaging_upstream_tip)
+        self.assertEqual(2, len(revtree.get_parent_ids()))
+        self.assertEqual(upstream_rev, revtree.get_parent_ids()[1])
+        # And the file has the new id in our tree
+        self.assertEqual("a-id", tree.path2id("a"))
+
+    def test_merge_upstream_with_dash_in_version_number(self):
+        tree = self.make_branch_and_tree('work')
+        self.build_tree(['work/a'])
+        tree.add(['a'])
+        orig_upstream_rev = tree.commit("one")
+        tree.branch.tags.set_tag("upstream-0.1", orig_upstream_rev)
+        self.build_tree(['work/debian/'])
+        cl = self.make_changelog(version="0.1-1")
+        self.write_changelog(cl, 'work/debian/changelog')
+        tree.add(['debian/', 'debian/changelog'])
+        orig_debian_rev = tree.commit("two")
+        db = DistributionBranch(tree.branch, None, tree=tree)
+        dbs = DistributionBranchSet()
+        dbs.add_branch(db)
+        tarball_filename = "package-0.2.tar.gz"
+        tf = tarfile.open(tarball_filename, 'w:gz')
+        try:
+            f = open("a", "wb")
+            try:
+                f.write("aaa")
+            finally:
+                f.close()
+            tf.add("a")
+        finally:
+            tf.close()
+        conflicts = db.merge_upstream(tarball_filename, "0.2-1", "0.1")
+        # Check that we tagged wiht the dash version
+        self.assertTrue(tree.branch.tags.has_tag('upstream-0.2-1'))
+
+    def test_merge_upstream_rename_and_replace(self):
+        """Test renaming a file upstream and replacing it.
+
+        We want to take the rename in to our tree, but have to be
+        careful not to assign the file id to the new file at the same
+        path as well, as that will lead to problems.
+        """
+        self.requireFeature(PristineTarFeature)
+        version1 = Version("1.0-1")
+        version2 = Version("1.1-1")
+        upstream_tree = self.make_branch_and_tree("upstream")
+        upstream_tree.lock_write()
+        self.addCleanup(upstream_tree.unlock)
+        self.build_tree(['upstream/a'])
+        upstream_tree.add(['a'], ['a-id'])
+        upstream_rev1 = upstream_tree.commit("one")
+        tree = upstream_tree.bzrdir.sprout('packaging').open_workingtree()
+        db = DistributionBranch(tree.branch, None, tree=tree)
+        dbs = DistributionBranchSet()
+        dbs.add_branch(db)
+        tree.lock_write()
+        self.addCleanup(tree.unlock)
+        tree.commit("add packaging")
+        tree.branch.tags.set_tag("upstream-%s" % version1.upstream_version,
+                upstream_rev1)
+        builder = SourcePackageBuilder("package", version2)
+        builder.add_default_control()
+        builder.add_upstream_file("a", "New a")
+        builder.add_upstream_file("b", "Renamed a")
+        builder.build()
+        upstream_tree.rename_one('a', 'b')
+        # We don't add the new file upstream, as the new file id would
+        # be picked up from there.
+        upstream_rev2 = upstream_tree.commit("two")
+        db.merge_upstream(builder.tar_name(), version2.upstream_version,
+            version1.upstream_version,
+            upstream_branch=upstream_tree.branch,
+            upstream_revision=upstream_rev2)
+        self.assertEqual("a-id", tree.path2id("b"))
+
+    def test_merge_upstream_rename_on_top(self):
+        """Test renaming a file upstream, replacing an existing file."""
+        self.requireFeature(PristineTarFeature)
+        version1 = Version("1.0-1")
+        version2 = Version("1.1-1")
+        upstream_tree = self.make_branch_and_tree("upstream")
+        upstream_tree.lock_write()
+        self.addCleanup(upstream_tree.unlock)
+        self.build_tree(['upstream/a', 'upstream/b'])
+        upstream_tree.add(['a', 'b'], ['a-id', 'b-id'])
+        upstream_rev1 = upstream_tree.commit("one")
+        tree = upstream_tree.bzrdir.sprout('packaging').open_workingtree()
+        db = DistributionBranch(tree.branch, None, tree=tree)
+        dbs = DistributionBranchSet()
+        dbs.add_branch(db)
+        tree.lock_write()
+        self.addCleanup(tree.unlock)
+        tree.commit("add packaging")
+        tree.branch.tags.set_tag("upstream-%s" % version1.upstream_version,
+                upstream_rev1)
+        builder = SourcePackageBuilder("package", version2)
+        builder.add_default_control()
+        builder.add_upstream_file("b", "Renamed a")
+        builder.build()
+        upstream_tree.unversion(['b-id'])
+        os.unlink('upstream/b')
+        upstream_tree.rename_one('a', 'b')
+        # We don't add the new file upstream, as the new file id would
+        # be picked up from there.
+        upstream_rev2 = upstream_tree.commit("two")
+        db.merge_upstream(builder.tar_name(), version2.upstream_version,
+            version1.upstream_version,
+            upstream_branch=upstream_tree.branch,
+            upstream_revision=upstream_rev2)
+        self.assertEqual("a-id", tree.path2id("b"))
+
     def test_import_symlink(self):
         version = Version("1.0-1")
         self.requireFeature(PristineTarFeature)
@@ -1708,3 +1888,21 @@ class SourceExtractorTests(tests.TestCaseInTempDir):
             self.assertTrue(os.path.exists(extractor.unextracted_upstream))
         finally:
             extractor.cleanup()
+
+
+class PossibleUpstreamTagNamesTests(BuilddebTestCase):
+
+    def setUp(self):
+        super(PossibleUpstreamTagNamesTests, self).setUp()
+        self.tree = self.make_branch_and_tree('unstable')
+        root_id = self.tree.path2id("")
+        self.up_tree = self.make_branch_and_tree('unstable-upstream')
+        self.up_tree.set_root_id(root_id)
+        self.db = DistributionBranch(self.tree.branch,
+                self.up_tree.branch, tree=self.tree,
+                upstream_tree=self.up_tree)
+
+    def test_version(self):
+        self.assertEquals(['upstream-3.3', 'upstream-debian-3.3',
+            'upstream-ubuntu-3.3', 'upstream/3.3'],
+            self.db.possible_upstream_tag_names("3.3"))
