@@ -18,7 +18,11 @@
 #    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 #
 
-from debian_bundle.changelog import Version
+try:
+    from debian.changelog import Version
+except ImportError:
+    # Prior to 0.1.15 the debian module was called debian_bundle
+    from debian_bundle.changelog import Version
 
 from bzrlib.revision import Revision
 from bzrlib.tests import TestCase, TestCaseWithTransport
@@ -144,17 +148,17 @@ class TestUpstreamTagToVersion(TestCase):
 class TestPackageVersion(TestCase):
 
   def test_simple_debian(self):
-    self.assertEquals(Version("1.2-1"), 
-        package_version(Version("1.2"), "debian"))
+    self.assertEquals(Version("1.2-1"),
+        package_version("1.2", "debian"))
 
   def test_simple_ubuntu(self):
-    self.assertEquals(Version("1.2-0ubuntu1"), 
-        package_version(Version("1.2"), "ubuntu"))
+    self.assertEquals(Version("1.2-0ubuntu1"),
+        package_version("1.2", "ubuntu"))
 
-  def test_debian_merges_ubuntu(self):
-    self.assertEquals(Version("1.2-1"), 
-        package_version(Version("1.2-0ubuntu1"), "debian"))
+  def test_debian_with_dash(self):
+    self.assertEquals(Version("1.2-0ubuntu1-1"),
+        package_version("1.2-0ubuntu1", "debian"))
 
-  def test_ubuntu_merges_debian(self):
-    self.assertEquals(Version("1.2-1ubuntu1"), 
-        package_version(Version("1.2-1"), "ubuntu"))
+  def test_ubuntu_with_dash(self):
+    self.assertEquals(Version("1.2-1-0ubuntu1"),
+        package_version("1.2-1", "ubuntu"))
