@@ -25,10 +25,17 @@ except ImportError:
 import os
 import shutil
 
-from debian_bundle.changelog import Changelog, Version
+try:
+    from debian.changelog import Changelog, Version
+except ImportError:
+    # Prior to 0.1.15 the debian module was called debian_bundle
+    from debian_bundle.changelog import Changelog, Version
 
+from bzrlib.plugins.builddeb.config import DebBuildConfig
 from bzrlib.plugins.builddeb.errors import (MissingChangelogError,
                 AddChangelogError,
+                NoPreviousUpload,
+                UnknownDistribution,
                 )
 from bzrlib.plugins.builddeb.tests import SourcePackageBuilder
 from bzrlib.plugins.builddeb.util import (
@@ -38,7 +45,9 @@ from bzrlib.plugins.builddeb.util import (
                   find_changelog,
                   find_extra_authors,
                   find_last_distribution,
+                  _find_previous_upload,
                   find_thanks,
+                  get_export_upstream_revision,
                   get_commit_info_from_changelog,
                   get_snapshot_revision,
                   lookup_distribution,
@@ -698,3 +707,68 @@ class FindLastDistributionTests(TestCase):
         changelog = self.create_changelog("UNRELEASED")
         self.assertEquals(None, find_last_distribution(changelog))
 
+
+class GetExportUpstreamRevisionTests(TestCase):
+
+    def test_snapshot_rev(self):
+        config = DebBuildConfig([])
+        self.assertEquals("34", 
+            get_export_upstream_revision(config, Version("0.1+bzr34-1")))
+
+    def test_export_upstream_rev(self):
+        config = DebBuildConfig([
+            ({"BUILDDEB": {"export-upstream-revision": "tag:foobar"}}, True)])
+        self.assertEquals("tag:foobar", 
+            get_export_upstream_revision(config, Version("0.1-1")))
+
+    def test_export_upstream_rev_var(self):
+        config = DebBuildConfig([({"BUILDDEB": 
+            {"export-upstream-revision": "tag:foobar-$UPSTREAM_VERSION"}},
+            True)])
+        self.assertEquals("tag:foobar-0.1", 
+            get_export_upstream_revision(config, Version("0.1-1")))
+
+    def test_export_upstream_rev_not_set(self):
+        config = DebBuildConfig([])
+        self.assertEquals(None, 
+            get_export_upstream_revision(config, Version("0.1-1")))
+
+
+class FindPreviousUploadTests(TestCase):
+
+    def make_changelog(self, versions_and_distributions):
+        cl = Changelog()
+        changes = ["  [ A. Hacker ]", "  * Something"]
+        author = "J. Maintainer <maint@example.com>"
+        for version, distro in versions_and_distributions:
+            cl.new_block(changes=changes, author=author,
+                                distributions=distro, version=version)
+        return cl
+
+    def test_find_previous_upload_debian(self):
+        cl = self.make_changelog([("0.1-1", "unstable"),
+                ("0.1-2", "unstable")])
+        self.assertEqual(Version("0.1-1"), _find_previous_upload(cl))
+        cl = self.make_changelog([("0.1-1", "unstable"),
+                ("0.1-1.1", "stable-security"), ("0.1-2", "unstable")])
+        self.assertEqual(Version("0.1-1"), _find_previous_upload(cl))
+
+    def test_find_previous_upload_ubuntu(self):
+        cl = self.make_changelog([("0.1-1", "lucid"),
+                ("0.1-2", "lucid")])
+        self.assertEqual(Version("0.1-1"), _find_previous_upload(cl))
+        cl = self.make_changelog([("0.1-1", "lucid"),
+                ("0.1-1.1", "unstable"), ("0.1-2", "maverick")])
+        self.assertEqual(Version("0.1-1"), _find_previous_upload(cl))
+
+    def test_find_previous_upload_unknown(self):
+        cl = self.make_changelog([("0.1-1", "lucid"),
+                ("0.1-2", "UNRELEASED")])
+        self.assertRaises(UnknownDistribution, _find_previous_upload, cl)
+
+    def test_find_previous_upload_missing(self):
+        cl = self.make_changelog([("0.1-1", "unstable"),
+                ("0.1-2", "lucid")])
+        self.assertRaises(NoPreviousUpload, _find_previous_upload, cl)
+        cl = self.make_changelog([("0.1-1", "unstable")])
+        self.assertRaises(NoPreviousUpload, _find_previous_upload, cl)
