@@ -30,14 +30,21 @@ import re
 
 from bzrlib.trace import mutter
 
-from debian_bundle import deb822
-from debian_bundle.changelog import Changelog, ChangelogParseError
+try:
+    from debian import deb822
+    from debian.changelog import Changelog, ChangelogParseError
+except ImportError:
+    # Prior to 0.1.15 the debian module was called debian_bundle
+    from debian_bundle import deb822
+    from debian_bundle.changelog import Changelog, ChangelogParseError
 
 from bzrlib import (
         bugtracker,
         errors,
         urlutils,
+        version_info as bzr_version_info,
         )
+from bzrlib.export import export as bzr_export
 from bzrlib.trace import warning
 from bzrlib.transport import (
     do_catching_redirections,
@@ -52,10 +59,22 @@ from bzrlib.plugins.builddeb.config import DebBuildConfig
 from bzrlib.plugins.builddeb.errors import (
                 MissingChangelogError,
                 AddChangelogError,
+                NoPreviousUpload,
+                UnableToFindPreviousUpload,
+                UnknownDistribution,
                 UnparseableChangelog,
                 )
 
 
+DEBIAN_RELEASES = ('woody', 'sarge', 'etch', 'lenny', 'squeeze', 'stable',
+        'testing', 'unstable', 'experimental', 'frozen', 'sid')
+DEBIAN_POCKETS = ('', '-security', '-proposed-updates', '-backports')
+UBUNTU_RELEASES = ('warty', 'hoary', 'breezy', 'dapper', 'edgy',
+        'feisty', 'gutsy', 'hardy', 'intrepid', 'jaunty', 'karmic',
+        'lucid', 'maverick')
+UBUNTU_POCKETS = ('', '-proposed', '-updates', '-security', '-backports')
+
+
 def safe_decode(s):
     """Decode a string into a Unicode value."""
     if isinstance(s, unicode): # Already unicode
@@ -109,7 +128,8 @@ def find_changelog(t, merge, max_blocks=1):
 
     :param t: the Tree to look in.
     :param merge: whether this is a "merge" package.
-    :param max_blocks: Number of max_blocks to parse (defaults to 1)
+    :param max_blocks: Number of max_blocks to parse (defaults to 1). Use None
+        to parse the entire changelog.
     :return: (changelog, larstiq) where changelog is the Changelog,
         and larstiq is a boolean indicating whether the file is at
         'changelog' if merge was given, False otherwise.
@@ -239,15 +259,8 @@ def suite_to_distribution(suite):
     :param suite: the string containing the suite
     :return: "debian", "ubuntu", or None if the distribution couldn't be inferred.
     """
-    debian_releases = ('woody', 'sarge', 'etch', 'lenny', 'squeeze', 'stable',
-            'testing', 'unstable', 'experimental', 'frozen', 'sid')
-    debian_targets = ('', '-security', '-proposed-updates', '-backports')
-    ubuntu_releases = ('warty', 'hoary', 'breezy', 'dapper', 'edgy',
-            'feisty', 'gutsy', 'hardy', 'intrepid', 'jaunty', 'karmic',
-            'lucid', 'maverick')
-    ubuntu_targets = ('', '-proposed', '-updates', '-security', '-backports')
-    all_debian = [r + t for r in debian_releases for t in debian_targets]
-    all_ubuntu = [r + t for r in ubuntu_releases for t in ubuntu_targets]
+    all_debian = [r + t for r in DEBIAN_RELEASES for t in DEBIAN_POCKETS]
+    all_ubuntu = [r + t for r in UBUNTU_RELEASES for t in UBUNTU_POCKETS]
     if suite in all_debian:
         return "debian"
     if suite in all_ubuntu:
@@ -267,18 +280,24 @@ def lookup_distribution(distribution_or_suite):
     return suite_to_distribution(distribution_or_suite)
 
 
+def md5sum_filename(filename):
+    """Calculate the md5sum of a file by name."""
+    m = md5.md5()
+    f = open(filename, 'rb')
+    try:
+        for line in f:
+            m.update(line)
+    finally:
+        f.close()
+    return m.hexdigest()
+
+
 def move_file_if_different(source, target, md5sum):
     if os.path.exists(target):
         if os.path.samefile(source, target):
             return
-        t_md5sum = md5.md5()
-        target_f = open(target)
-        try:
-            for line in target_f:
-                t_md5sum.update(line)
-        finally:
-            target_f.close()
-        if t_md5sum.hexdigest() == md5sum:
+        t_md5sum = md5sum_filename(target)
+        if t_md5sum == md5sum:
             return
     shutil.move(source, target)
 
@@ -477,7 +496,7 @@ def find_last_distribution(changelog):
     """Find the last changelog that was used in a changelog.
 
     This will skip stanzas with the 'UNRELEASED' distribution.
-    
+
     :param changelog: Changelog to analyze
     """
     for block in changelog._blocks:
@@ -494,12 +513,11 @@ def subprocess_setup():
     signal.signal(signal.SIGPIPE, signal.SIG_DFL)
 
 
-def debuild_config(tree, working_tree, no_user_config):
+def debuild_config(tree, working_tree):
     """Obtain the Debuild configuration object.
 
     :param tree: A Tree object, can be a WorkingTree or RevisionTree.
     :param working_tree: Whether the tree is a working tree.
-    :param no_user_config: Whether to skip the user configuration
     """
     config_files = []
     user_config = None
@@ -509,12 +527,85 @@ def debuild_config(tree, working_tree, no_user_config):
                         "local.conf"))
         else:
             warning('Not using configuration from %s as it is versioned.')
-    if not no_user_config:
-        config_files.append((global_conf, True))
-        user_config = global_conf
+    config_files.append((global_conf(), True))
+    user_config = global_conf()
     if tree.path2id(default_conf):
         config_files.append((tree.get_file(tree.path2id(default_conf)), False,
                     "default.conf"))
     config = DebBuildConfig(config_files, tree=tree)
     config.set_user_config(user_config)
     return config
+
+
+def export(tree, dest, format=None, root=None, subdir=None, filtered=False,
+           require_per_file_timestamps=False):
+    """Simple wrapper around bzrlib.export.export that prefers 
+    per_file_timestamps if it is supported.
+
+    """
+    # per_file_timestamps is available as of bzr 2.2.0
+    if bzr_version_info >= (2, 2, 0):
+        return bzr_export(tree, dest, format=format, root=root, subdir=subdir,
+            filtered=filtered, per_file_timestamps=True)
+    else:
+        if require_per_file_timestamps:
+            raise errors.PerFileTimestampsNotSupported()
+        return bzr_export(tree, dest, format=format, root=root, subdir=subdir,
+            filtered=filtered)
+
+
+def find_previous_upload(tree, merge):
+    """Given a tree, find the previous upload to the distribution.
+
+    When e.g. Ubuntu merges from Debian they want to build with
+    -vPREV_VERSION. Here's where we find that previous version.
+
+    We look at the last changelog entry and find the upload target.
+    We then search backwards until we find the same target. That's
+    the previous version that we return.
+
+    We require there to be a previous version, otherwise we throw
+    an error.
+
+    It's not a simple string comparison to find the same target in
+    a previous version, as we should consider old series in e.g.
+    Ubuntu.
+    """
+    try:
+        cl, larstiq = find_changelog(tree, merge, max_blocks=None)
+    except UnparseableChangelog:
+        raise UnableToFindPreviousUpload()
+    return _find_previous_upload(cl)
+
+
+def _find_previous_upload(cl):
+    blocks = cl._blocks
+    current_target = blocks[0].distributions.split(" ")[0]
+    all_debian = [r + t for r in DEBIAN_RELEASES for t in DEBIAN_POCKETS]
+    all_ubuntu = [r + t for r in UBUNTU_RELEASES for t in UBUNTU_POCKETS]
+    if current_target in all_debian:
+        match_targets = (current_target,)
+    elif current_target in all_ubuntu:
+        match_targets = UBUNTU_RELEASES
+        if "-" in current_target:
+            match_targets += [current_target.split("-", 1)[0]
+                + t for t in UBUNTU_POCKETS]
+    else:
+        raise UnknownDistribution(current_target)
+    previous_version = None
+    for block in blocks[1:]:
+        if block.distributions.split(" ")[0] in match_targets:
+            return block.version
+    raise NoPreviousUpload(current_target)
+
+
+def tree_contains_upstream_source(tree):
+    """Guess if the specified tree contains the upstream source.
+    
+    :param tree: A RevisionTree.
+    :return: Boolean indicating whether or not the tree contains the upstream
+        source
+    """
+    present_files = set(tree.inventory.root.children.keys())
+    packaging_files = frozenset(["debian", ".bzr-builddeb"])
+    return (len(present_files - packaging_files) > 0)
